--[[

TPT All script
LBPHacker, 2018

This script lets you define sets of particles, filter these sets (comparisons,
user filters), execute boolean operations on them (union, difference, intersection)
and finally manipulate the particles in them en masse. It also lets you access
properties of particles through proxy objects.

DO READ THIS. It's worth it. I mean it. You won't know how to use this if you don't.

It does assume a bit of Lua knowledge but Lua is a language filled to the brim
with easy to discover patterns, so don't be afraid of experimenting. The worst
that can happen is you get an error.

The script sets up a table in the global tpt table, it's called tpt.all. All
functions defined by the script are in tpt.all.

There are two main concepts in this script that you should wrap your head around:
sets and wrappers. Sets are, well, sets of wrappers, and wrappers are proxies
around individual particles. You can use these wrappers to access properties of
particles easily (i.e. not with sim.partProperty). 

Wrappers around particles are returned by tpt.all.wrapper. It expects a particle
ID as a parameter and returns an object that you can index with property names.
For performance reasons, no sanity checking is done when accessing a property through
such a wrapper, so if you try to access a property that doesn't exist, you may get
an error. Example:

	local wrapper = tpt.all.wrapper(1337)
	wrapper.ctype = 0xBADC0DE -- equivalent to sim.partProperty(1337, "ctype", 0xBADC0DE)
	print(wrapper.life) -- equivalent to print(sim.partProperty(1337, "life"))

Note a wrapper is dumb. If the particle it wraps dies, it won't know about it
and if you try to access a dead particle through its wrapper, you may get an error.

Again, no sanity checking is done. For this reason it's best to use the functions
in this script while the simulation is paused. To be honest, I can't think of a
reason to use them when it's not paused, but I'm sure the internet will find one.

The script makes sure that at any given moment only one wrapper exists for any
particle ID. This means that if you call tpt.all.wrapper multiple times with the same
ID, you'll get the same object back. This makes sets easy to implement.

Sets are more fun. They are actually sets of wrappers but let's just pretend that
they are sets of particles. There are many ways to create such sets:

- tpt.all() returns a set of all particles in the simulation,
- tpt.assoc(tbl) returns a set of particles whose IDs are the keys of tbl,
- tpt.array(tbl) returns a set of particles whose IDs are the values of tbl (a Lua array),
- tpt.buffer(func, ...) returns a set of particles created by func (a function which
  the script calls with the rest of the arguments, the ... thing),
- tpt.one(id) returns a set which holds a single particle whose id is id.

In fact you shouldn't use tpt.all.wrapper at all. Sets are way better, believe me.

Now that you have sets, it's time to manipulate the particles in them a bit. Here are
a bunch of examples of such en masse manipulations:

	tpt.all():set("dcolour", 0xFFFF0000) -- paint all particles red
	tpt.all():set("ctype", 0xBADC0DE)
	print(tpt.all:count())
	tpt.all():kill()

--]]

local function to_property_value(property_value)
	if type(property_value) == "string" then
		local elementID = elem["DEFAULT_PT_" .. property_value:upper()]
		if elementID then
			property_value = elementID
		end
	end
	return property_value
end

local particle_wrapper_m = {}
function particle_wrapper_m:__index(property_key)
	return sim.partProperty(self.id, property_key)
end
function particle_wrapper_m:__newindex(property_key, property_value)
	return sim.partProperty(self.id, property_key, to_property_value(property_value))
end

local weak_wrapper_store = setmetatable({}, {__mode = "v"})
local function make_particle_wrapper(id)
	local grab_wrapper = weak_wrapper_store[id]
	if grab_wrapper then
		return grab_wrapper
	end
	local new_wrapper = setmetatable({
		id = id
	}, particle_wrapper_m)
	weak_wrapper_store[id] = new_wrapper
	return new_wrapper
end

local make_particle_set

local particle_set_i = {}
local particle_set_m = {__index = particle_set_i}

function particle_set_m:__add(other)
	local result = make_particle_set()
	for id in self:iterate() do
		result:insert(id)
	end
	for id in other:iterate() do
		result:insert(id)
	end
	return result
end
function particle_set_m:__sub(other)
	local result = make_particle_set()
	for id in self:iterate() do
		result:insert(id)
	end
	for id in other:iterate() do
		result:remove(id)
	end
	return result
end
function particle_set_m:__mul(other)
	local result = make_particle_set()
	for id in self:iterate() do
		result:insert(id)
	end
	for id in (tpt.all() - other):iterate() do
		result:remove(id)
	end
	return result
end
function particle_set_m:__eq(other)
	if self:count() ~= other:count() then
		return false
	end
	for id in self:iterate() do
		if not other:has(id) then
			return false
		end
	end
	return true
end

function particle_set_i:next(current_id)
	local next_id, next_wrapper = next(self.particle_assoc, current_id)
	if not next_id then
		return
	end
	return next_id, next_wrapper
end

function particle_set_i:has(id)
	return self.particle_assoc[id] and true
end

function particle_set_i:iterate()
	return self.next, self
end

function particle_set_i:bbox(x1, y1, x2, y2)
	return self:gte("x", x1):gte("y", y1):lte("x", x2):lte("y", y2)
end

function particle_set_i:filter(do_insert_func)
	local result = make_particle_set()
	for id, wrapper in self:iterate() do
		if do_insert_func(wrapper) then
			result:insert(id)
		end
	end
	return result
end

function particle_set_i:copy()
	return self:filter(function()
		return true
	end)
end

function particle_set_i:eq(property_key, property_value)
	property_value = to_property_value(property_value)
	return self:filter(function(wrapper)
		return wrapper[property_key] == property_value
	end)
end

function particle_set_i:neq(property_key, property_value)
	property_value = to_property_value(property_value)
	return self:filter(function(wrapper)
		return wrapper[property_key] ~= property_value
	end)
end

function particle_set_i:gt(property_key, property_value)
	property_value = to_property_value(property_value)
	return self:filter(function(wrapper)
		return wrapper[property_key] > property_value
	end)
end

function particle_set_i:gte(property_key, property_value)
	property_value = to_property_value(property_value)
	return self:filter(function(wrapper)
		return wrapper[property_key] >= property_value
	end)
end

function particle_set_i:lt(property_key, property_value)
	property_value = to_property_value(property_value)
	return self:filter(function(wrapper)
		return wrapper[property_key] < property_value
	end)
end

function particle_set_i:lte(property_key, property_value)
	property_value = to_property_value(property_value)
	return self:filter(function(wrapper)
		return wrapper[property_key] <= property_value
	end)
end

function particle_set_i:each(func)
	for id, wrapper in self:iterate() do
		func(wrapper)
	end
	return self
end

function particle_set_i:set(property_key, property_value)
	property_value = to_property_value(property_value)
	if type(property_value) == "function" then
		for id, wrapper in self:iterate() do
			wrapper[property_key] = property_value(wrapper)
		end
	else
		for id, wrapper in self:iterate() do
			wrapper[property_key] = property_value
		end
	end
	return self
end

function particle_set_i:kill()
	return self:each(function(wrapper)
		sim.partKill(wrapper.id)
	end)
end

function particle_set_i:get(property_key)
	local result
	local multiple_values = false
	for id, wrapper in self:iterate() do
		local value = tostring(wrapper[property_key])
		if result then
			if result ~= value then
				multiple_values = true
				break
			end
		else
			result = value
		end
	end
	return multiple_values and "[multiple values]" or result
end

function particle_set_i:insert(id)
	if not self.particle_assoc[id] then
		self.particle_assoc[id] = make_particle_wrapper(id)
		self.particle_count = self.particle_count + 1
	end
	return self
end

function particle_set_i:remove(id)
	if self.particle_assoc[id] then
		self.particle_assoc[id] = nil
		self.particle_count = self.particle_count - 1
	end
	return self
end

function particle_set_i:count()
	return self.particle_count
end

function make_particle_set()
	return setmetatable({
		particle_count = 0,
		particle_assoc = {}
	}, particle_set_m)
end

tpt.all = setmetatable({}, {__call = function(self, param, param2)
	if type(param2) == "number" and type(param) == "number" then
		return tpt.all.at(param, param2)
	end
	if type(param) == "number" then
		return tpt.all.wrapper(param)
	elseif type(param) == "table" then
		return tpt.all.array(param)
	end
	local all_particles = make_particle_set()
	for id in sim.parts() do
		all_particles:insert(id)
	end
	return all_particles
end})

function tpt.all.at(x, y)
	local id = sim.partID(x, y)
	return id and tpt.all.wrapper(id) or nil
end

function tpt.all.wrapper(id)
	return make_particle_wrapper(id)
end

function tpt.all.one(id)
	local one_particle = make_particle_set()
	one_particle:insert(id)
	return one_particle
end

function tpt.all.assoc(id_assoc)
	local all_particles = make_particle_set()
	for id in pairs(id_assoc) do
		all_particles:insert(id)
	end
	return all_particles
end

function tpt.all.array(id_array)
	local all_particles = make_particle_set()
	for ix = 1, #id_array do
		all_particles:insert(id_array[ix])
	end
	return all_particles
end

function tpt.all.buffer(func, ...)
	local id_buffer = {}
	local sim_partCreate_original = sim.partCreate
	sim.partCreate = function(...)
		local result = {pcall(sim_partCreate_original, ...)}
		if not result[1] then
			error(result[2], 2)
		end
		if result[2] then
			id_buffer[result[2]] = true
		end
		return unpack(result, 2)
	end
	func(...)
	local result = tpt.all.assoc(id_buffer)
	sim.partCreate = sim_partCreate_original
	return result
end

